/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.    
 */

package org.apache.tuscany.sca.implementation.java.invocation;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.apache.tuscany.sca.assembly.ConfiguredOperation;
import org.apache.tuscany.sca.assembly.Contract;
import org.apache.tuscany.sca.assembly.OperationsConfigurator;
import org.apache.tuscany.sca.implementation.java.JavaImplementation;
import org.apache.tuscany.sca.invocation.Interceptor;
import org.apache.tuscany.sca.invocation.InvocationChain;
import org.apache.tuscany.sca.invocation.Invoker;
import org.apache.tuscany.sca.policy.Intent;
import org.apache.tuscany.sca.policy.PolicySet;
import org.apache.tuscany.sca.policy.PolicySetAttachPoint;
import org.apache.tuscany.sca.policy.util.PolicyHandler;
import org.apache.tuscany.sca.policy.util.PolicyHandlerTuple;
import org.apache.tuscany.sca.runtime.RuntimeComponent;
import org.apache.tuscany.sca.runtime.RuntimeComponentReference;
import org.apache.tuscany.sca.runtime.RuntimeWire;
import org.apache.tuscany.sca.runtime.RuntimeWireProcessor;

/**
 * Processor to inject policy handling interceptor wheneve PolicySets are specified in a Java Implementation 
 */
public class JavaPolicyHandlingRuntimeWireProcessor implements RuntimeWireProcessor {
    public JavaPolicyHandlingRuntimeWireProcessor() {
        super();
    }

    public void process(RuntimeWire wire) {
        /*Contract contract = wire.getSource().getContract();
        if (!(contract instanceof RuntimeComponentReference)) {
            return;
        }*/
        
        RuntimeComponent component = wire.getTarget().getComponent();
        if ( component != null && component.getImplementation() instanceof JavaImplementation ) {
            JavaImplementation javaImpl = (JavaImplementation)component.getImplementation();
            if ( javaImpl instanceof PolicySetAttachPoint ) {
                PolicyHandler policyHandler = null;
                List<PolicyHandler> implPolicyHandlers = new ArrayList<PolicyHandler>();
                PolicySetAttachPoint policiedImpl = (PolicySetAttachPoint)javaImpl;
                
                try {
                    //for ( PolicySet policySet : policiedImpl.getPolicySets() ) {
                    for ( PolicySet policySet : component.getPolicySets() ) {
                        policyHandler = getPolicyHandler(policySet, javaImpl.getPolicyHandlerClassNames());
                        if ( policyHandler != null ) {
                            policyHandler.setUp(javaImpl);
                            implPolicyHandlers.add(policyHandler);
                        } else {
                            //FIXME: to be removed after the PolicyHandler story has crystalized..
                            //maybe replace with exception then...
                            System.out.println("No PolicyHandler registered for PolicySet - " + policySet.getName());
                        }
                    }    
                    
                    List<PolicyHandler> applicablePolicyHandlers = null;
                    for (InvocationChain chain : wire.getInvocationChains() ) {
                        applicablePolicyHandlers = new ArrayList<PolicyHandler>();
                        if ( javaImpl instanceof OperationsConfigurator ) {
                            String operationName = chain.getTargetOperation().getName();
                            OperationsConfigurator opConfigurator = (OperationsConfigurator)component;
                            for ( ConfiguredOperation confOp : opConfigurator.getConfiguredOperations() ) {
                                if ( confOp.getName().equals(operationName)) {
                                    for ( PolicySet policySet : confOp.getPolicySets() ) {
                                        policyHandler = getPolicyHandler(policySet, javaImpl.getPolicyHandlerClassNames());
                                        policyHandler.setUp(javaImpl);
                                        applicablePolicyHandlers.add(policyHandler);
                                    }
                                    break;
                                }
                            }
                            
                            //if no policies have been specified at the operation level then simply
                            //apply whatever is specified for the implementation level
                            if ( applicablePolicyHandlers.isEmpty() ) {
                                applicablePolicyHandlers = implPolicyHandlers;
                            }
                        }
                        
                        if ( !applicablePolicyHandlers.isEmpty() ) {
                            int index = 0;
                            Invoker invoker = chain.getHeadInvoker();
                            while ( invoker != chain.getTailInvoker()) {
                                if ( invoker instanceof Interceptor ) {
                                    invoker = ((Interceptor)invoker).getNext();
                                    ++index;
                                }
                            }
                            chain.addInterceptor(index, new PolicyHandlingInterceptor(chain.getTargetOperation(),
                                                                                  applicablePolicyHandlers));
                        }
                    }
                } catch ( Exception e ) {
                    throw new RuntimeException(e);
                }
            }
        }
    }
    
    private PolicyHandler getPolicyHandler(PolicySet policySet, 
                                           Map<ClassLoader, List<PolicyHandlerTuple>> policyHandlerClassNames) 
                                throws IllegalAccessException, ClassNotFoundException, InstantiationException {
        
        PolicyHandler handler = null;
        
        for (ClassLoader classLoader : policyHandlerClassNames.keySet()) {
            for ( PolicyHandlerTuple handlerTuple : policyHandlerClassNames.get(classLoader) ) {
                for ( Intent intent : policySet.getProvidedIntents() ) {
                    if ( intent.getName().equals(handlerTuple.getProvidedIntentName())) {
                        for ( Object policy : policySet.getPolicies() ) {
                            if ( policy.getClass().getName().equals(handlerTuple.getPolicyModelClassName())) {
                                handler = 
                                    (PolicyHandler)Class.forName(handlerTuple.getPolicyHandlerClassName(), 
                                                                 true, 
                                                                 classLoader).newInstance();
                                    handler.setApplicablePolicySet(policySet);
                                    return handler;
                            }
                        }
                    }
                }
            }
        }
        
        return handler;
    }
                            
}
